# =====================================================================
# pyx2csa.awk: process RPC input variables from  a PYX request file.
#
# Copyright (c) 2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: pyx2csa.awk,v 1.2 2003/05/07 09:11:26 carlo Exp $

# =====================================================================
# Notes:
#
# Note that patterns include an optional namespace specification on
# the left-hand side of each tag, although namespaces are not yet
# supported by current XML-RPC specs.
#
# See http://www.xmlrpc.com/discuss/msgReader$1208 for more info on
# the proposed system.multicall specs.
#
# According to the XML-RPC specs, <struct> elements are *unordered*
# lists of {name,value} pairs. To cope with that, for each struct
# I set an extra variable containing the *ordered* list of the composing
# elements.
#
# Application-level authentication, if any, is expected to be in the
# first '<param>' element (the CSA "system block"), and is extracted
# if the 'auth' flag is set to '1':
#
# <params>
#  <param>
#   <value><array>
#     <data>
#      <value><struct>          <!-- first array element is a struct -->
#        <member>
#         <name>HTTP_AUTHORIZATION</name>
#         <value><string>BASIC c21pdGhqOm15cGFzcw==</string></value>
#        </member>
#        <member>
#         <name>CSA_RPC_ACTION</name>
#         <value><string>GET</string></value>
#        </member>
#       </struct></value>
#      <value>...</value>       <!-- second erray element -->
#      <value>...</value>       <!-- third array element -->
#     ...                       <!-- other array elements ... -->
#     </data>
#   </array></value>
#  </param>
# 
# Like every '<struct>', also the system block is an *unordered*
# list of name/value pairs. The authentication extraction logic
# accounts for such unordering.
# =====================================================================

BEGIN {

  # Get output file-name from env.
  o_file = ENVIRON["CSA_RPC_WWW"]

  if (auth != 1) auth = 0
}

# =====================================================================
# Common stuff. 
# =====================================================================

# This must come very first.
decrease		{ --depth; decrease=0 }

# '<string>' elements do not change the current depth level.
/^\(.*:?string$/	{ --depth }
/^\).*:?string$/	{ ++depth }

/^\(/			{ ++depth }

# The need for decreasing must be detected now, but it must be applied
# only upon the beginning of the next loop.

/^\)/			{ decrease=1 }

# =====================================================================
# Decide whether it is time to print the assignment.
# =====================================================================

/^\)/ && data != "" {

   invar(name append, data)

   if (append == "_N") {

      # Handle the system parameter block.

      if (name ~ /^WWW_1_1_1_[1-9][0-9]*$/) {
	 if (auth && data == "HTTP_AUTHORIZATION") authvar = name
	 else if (data == "CSA_RPC_ACTION") verbvar = name
      }

      sub(/_[0-9]+$/,"",name)
      if (list[name] == "") list[name] = data
      else list[name] = list[name] "\001" data
   }

   # Set authentication token, if requested.
   if (auth && append == "" && name == authvar) {
      authset(data, o_file)
      rcassign(authvar, "", o_file)		# Clear original var.
   }

   # Set virtual REST method.
   if (append == "" && name == verbvar && \
		data ~ /^(DELETE|GET|HEAD|POST|PUT)$/) {
      invar("CSA_RPC_ACTION",data)
      rcassign(verbvar, "", o_file)		# Clear original var.
   }

   data = ""
}

/^\(.*:?param$/  { id[++level]++ }
/^\).*:?param$/  { id[level+1]=0; level-- }

/^\(.*:?value$/  { id[++level]++; value=1 }
/^\).*:?value$/  { id[level+1]=0; level--; value=0 }

/^\(.*:?name$/	 { id[++level]++; value=1; append = "_N" }
/^\).*:?name$/	 { id[level--]--; value=0; append = "" }

# base64 must be handled specially.
/^\(.*:?base64$/ {

   value=2
   o_tmp = name "_F"
   print "csaMkTemp " o_tmp > o_file
   print "<<'END-OF-BASE64' mimencode -u -o $" o_tmp \
	 " || csaExit.fault 0003 E 'mimencode -u'" > o_file
   next
}

value==2 && /^-\\n$/ {
   printf("\n") > o_file
   next
}

value==2 && /^-/ {
   printf("%s", substr($0,2)) > o_file
   next
}

# Skip datatype specs.
/^\).*:?(string|int|i4|boolean|double|dateTime\.iso8601|nil)$/ { next }

/^\(.*:?(struct|array)$/ {

   value=0

   # Try and cope with clients that send '<value>\n<struct>' type
   # of things. The PYX file preserves the newline, but in this
   # case it should be discarded, I think.

   data = ""

   next
}

/^\).*:?base64$/ {
   value=0
   printf("\n%s\n", "END-OF-BASE64") > o_file
   next
}

value {
   name = "WWW_" id[1]
   for (i=2; i<=level; i++) name = name "_" id[i]
}

value==1 && /^-\\n$/ {
   data = data "\n"
   next
}

value==1 && /^-/ {
   data = data unescpyx(substr($0,2))
}

END {

   for (i in list) invar(i, list[i])

   # Make sure the output file is created even if there
   # are no assignments.

   printf("\n") > o_file
}

# =====================================================================
# Function section
# =====================================================================

# =====================================================================
# string unescpyx(string s)
#
# Un-escape PYX special sequences \t and \\
# Note: global variables, like RSTART and RLENGTH cannot be localised.
# =====================================================================

function unescpyx(s,		S,rstart,rlength) {

   while (match(s,/\\+t/)) {

      rstart=RSTART; rlength=RLENGTH

      if (!(rlength % 2))
	 S = S substr(s,1,rstart+rlength-3) "\t"
      else
	 S = S substr(s,1,rstart+rlength-1)

      s = substr(s,rstart+rlength)
   }

   s = S s
   S = ""

   while (match(s,/[\\][\\]/)) {

      rstart=RSTART; rlength=RLENGTH

      S = S substr(s,1,rstart)
      s = substr(s,rstart+2)
   }

   return S s
}

# =====================================================================
# End of pyx2csa.awk
# =====================================================================
