# =====================================================================
# xmlrpc-multicall.awk: split a 'system.multicall' request into its
# individual method calls.
#
# Copyright (c) 2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: xmlrpc-multicall.awk,v 1.3 2003/04/28 20:33:19 carlo Exp $

# =====================================================================
# Notes:
#
# Note that patterns include an optional namespace specification on
# the left-hand side of each tag, although namespaces are not yet
# supported by current XML-RPC specs.
#
# See http://www.xmlrpc.com/discuss/msgReader$1208 for more info on
# the proposed system.multicall specs.
#
# According to the XML-RPC specs, <struct> elements are *unordered*
# lists of {name,value} pairs. That means that I must expect to receive
# a multicall payload where the 'params' struct member comes before the
# 'methodName' member. And indeed that happens, for instance with the
# XML-RPC test service at http://www.dscpl.com.au/xmlrpc-debugger.php
# However, we need to receive the 'methodName' element first, so I
# introduced the 'method' switch in the following code, to buffer
# 'params' contents and print them out later, if they are received
# first.
#
# Each individual 'target' is assigned a unique target-id (a positive
# integer). Targets can be single items or groups of related items.
# =====================================================================

# =====================================================================
# Scan the PYX request file.
# =====================================================================

BEGIN {

  # Process options.
  if (out == "") o_file = STDOUT
  else o_file = out

  # Init vars.
  rpcfile = "/dev/null"
  allow_multi = ENVIRON["CSA_INSTALL"] "/etc/multicall.allow"

  depth = call = target = params = method = x = y = numcalls = 0

  # Se the max. no. of boxcarred calls.
  maxcalls = ENVIRON["CSA_RPC_MAXCALLS"]
  if (maxcalls == "") maxcalls = 0		# unlimited by default.

  pyx_last = split(ENVIRON["CSA_WORKFILES"], a, "\001")

  # Begin the multicall response block.
  printf("echo -n '<value><array><data>'\n") > o_file

  # Load the list of multicall-enabled classes.
  while (getline class[++i] < allow_multi > 0);
  close(allow_multi)
}

# =====================================================================
# Common stuff. 
# =====================================================================

# This must come very first.
decrease		{ --depth; decrease=0 }

# '<string>' elements do not change the current depth level.
/^\(.*:?string$/	{ --depth }
/^\).*:?string$/	{ ++depth }

/^\(/			{ ++depth }

# The need for decreasing must be detected now, but it must be applied
# only upon the beginning of the next loop.

/^\)/			{ decrease=1 }

# =====================================================================
# Detect call starting point.
# =====================================================================

/^\(.*:?struct$/ && depth==8 {
  call = 1
  method = 0
  delete hold
  params = 0
  x = y = 0
  next
}

# =====================================================================
# 1st target in group "100".
# =====================================================================

/^-methodName$/ && depth==10 && call {
  target = 100
  next
}

# =====================================================================
# 2nd target in group "100". This is the called method name.
# =====================================================================

/^-([a-zA-Z0-9][-_:a-zA-Z0-9]*\.?)+$/ && depth==10 && target==100 {
  rpc = substr($1,2,64)
  rpcappl = rpc; sub(/\..*$/,"",rpcappl)
  rpcname = rpc; sub(/^[^.]+\./,"",rpcname)

  if (maxcalls && ++numcalls > maxcalls) {
     # system.multicall RFC mandates that:
     # "Under no circumstances may the implementation return a
     # list of the wrong length". So I cannot simply drop further
     # requests, but I must produce a fault message for each of them.
     # This will not save processing on the part of the server,
     # but at least the client may want to modify its code and
     # not do so many useless requests. I know, this is an easy
     # DoS to exploit, but CSA_RPC_MAXSIZE may help with that.
     rpcappl = "system"
     rpcname = "maxMulticallExceeded"
  }
  else if (rpcappl == "system" && rpcname == "multicall") {
     # recursive calls to system.multicall not allowed.
     # Force a different method name, which will print
     # a "recursion not allowed" fault struct when invoked.

     rpcname = "noMulticallRecursion"
  }
  else {
     allowed = 0;
     for (i in class) {
	 # skip comments and empty lines.
	 if (class[i] ~ /^[ \t]*(#|$)/) continue
	 if (class[i] == rpcappl) {
	    allowed = 1
	    break
	 }
     }
     if (!allowed) {
	# not a multicall-enabled class.
	rpcappl = "system"
	rpcname = "noMulticallClass"
     }
  }

  newcall()

  printf("(methodCall\n(methodName\n") > pyx
  printf("-%s.%s\n",rpcappl,rpcname) > pyx
  printf(")methodName\n(params\n") > pyx
  params = 1

  # Flush the postponed buffer and free memory.
  while (++y <= x) printf("%s", hold[y]) > pyx

  method = 1					# Got methodName.

  target = 0					# Group 100 complete.
  next
}

# =====================================================================
# 1st target in group "200".
# =====================================================================

/^-params$/ && depth==10 && call {
  target = 200
  next
}

# =====================================================================
# 2nd target in group "200". This is the method parameter block.
# =====================================================================

# Print block prologue.
/^\(.*:?value$/ && depth==13 && target==200 {
  if (!method) hold[++x] = "(param\n(value\n"
  else printf("(param\n(value\n") > pyx
  next
}
# Print sub-block separator.
/^\).*:?value$/ && depth==13 && target==200 {
  if (!method) hold[++x] = ")value\n)param\n"
  else printf(")value\n)param\n") > pyx
  next
}
# Block printing condition.
depth > 12 && target==200 {
  if (!method) hold[++x] = $0 "\n"
  else print > pyx
  next
}
# Detect target group end condition.
/^\).*:?data$/ && depth==12 && target==200 {
  target = 0					# Group 200 complete.
  next
}

# =====================================================================
# Detect call ending point.
# =====================================================================

/^\).*:?struct$/ && depth==8 {
  call = 0
  if (!method) hold[++x] = ")params\n)methodCall\n"
  else printf(")params\n)methodCall\n") > pyx
  next
}

# Uncomment for debugging. This *must* come last.
#{ print depth, call, $0 > "/dev/stderr" }

END {

  # End the multicall response block.
  printf("echo -n '</data></array></value>'\n") > o_file
}

# =====================================================================
# Function section
# =====================================================================

function newcall() {

  close(pyx)
  pyx = ENVIRON["TMPDIR"] "/" ENVIRON["CSA_PID"] "." ++pyx_last ".tmp"

  # Make sure partial PYX files are removed on 'system.multicall' exit.
  printf("csaTrapFile %s\n", pyx) > o_file

  # Run individual methods in subshells. There are good reasons for this.
  printf("@{\ncsaSubshell\n") > o_file
  if (rpcappl ~ /^[a-zA-Z0-9][-_:a-zA-Z0-9]*$/ && \
     rpcname ~ /^[a-zA-Z0-9][-_.:a-zA-Z0-9]*$/) {
     printf("CSA=%s/.csa/%s/csa.rc\n",ENVIRON["HOME"], rpcappl) > o_file
     printf("CSA_PGM='%s'\n",rpcname) > o_file

     # The test is necessary, otherwise if the file is missing the
     # error msg will be embedded into the HTTP response message.

     if (stat(ENVIRON["HOME"] "/.csa/" rpcappl "/csa.rc"))
        print "CSA_PGM=multicall;csaExit.fault 0067 E" > o_file

     printf(". $CSA || {\n") > o_file
     printf("  csaExit.fault 0067 E\n") > o_file
     printf("}\n") > o_file

     # rpclib *must* be reloaded, as the inclusion of mainlib by
     # the individual $CSA profile has overridden necessary functions.

     printf("csaLoadLib xmlrpclib.rc || {\n") > o_file
     printf("  csaExit.fault 0008 E xmlrpclib.rc\n") > o_file
     printf("}\n") > o_file

     printf("CSA_RPC_PYX=%s\n", pyx) > o_file

     # Re-use the top-level file created by urlencoded.awk.
     #printf("csaMkTemp CSA_RPC_WWW\n") > o_file

     # Now call 'pyx2csa.awk', to extract method call arguments
     # from the relevant PYX file.

     # Does the application-level argument-processing AWK
     # library exist ? Otherwise use the CSA default library.
     # Note that the user program has no way to specify
     # additional libraries. If they are needed, they will
     # have to be statically included in the user-level library.
     # If the user-level library exists, it must define at least
     # invar() and outvar().

     if (!stat(ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk"))
	arg_proc = "mawk -f^(($CSA_INSTALL/lib/mainlib " ENVIRON["HOME"] "/.csa/" rpcappl "/lib/rpcio/" rpcname ")^.awk)"
     else
	arg_proc = "mawk -f^($CSA_INSTALL/lib/^((mainlib rpciolib)^.awk))"

     # This will write rc(1) assignments to CSA_RPC_WWW.
     printf("%s '-vauth=1' -f$CSA_RPC_AWK -f$CSA_INSTALL/lib/pyx2csa.awk $CSA_RPC_PYX || csaExit.fault 0003 E mawk\n",arg_proc) > o_file

     # At this point CSA_RPC_WWW should have been written.
     printf(". $CSA_RPC_WWW || csaExit.fault 0008 E $CSA_RPC_WWW\n") > o_file

     rpcfile = ENVIRON["HOME"] "/.csa/" rpcappl "/rpclib/" rpcname
  }
  else printf("csaExit.fault 0068 E %s\n",rpc) > o_file

  while (getline record < rpcfile > 0) print record > o_file
  if (close(rpcfile)) {
     # Requested method not found.
     printf("csaExit.fault 0067 E\n") > o_file
  }

  printf("}\n") > o_file		# Subshell end statement.
}

# =====================================================================
# End of xmlrpc-multicall.awk
# =====================================================================
