/* csa-escape: copy stdin to stdout, applying the requested escaping
   of special characters along the way. Of course this could be
   done just as well with a sed(1) script, but the latter would be
   much slower on large input, and could not cope easily with the
   transformation of newlines.

   Copyright (c) 2003,2004 Carlo Strozzi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


   Usage:  csa-escape [-z|-Z|-p|-x|-i|-u|-a|-r|-s|-q|-Q]

   where:

     '-a' stands for AWK
     '-i' stands for ISO
     '-p' stands for PYX
     '-r' stands for rc(1)
     '-s' stands for sh(1)
     '-u' stands for URI
     '-x' stands for XML
     '-z' stands for 'binary zeroes', w/ trailing newline.
     '-Z' stands for 'binary zeroes', w/o trailing newline.
     '-q' stands for 'binary zeroes', w/ trailing newline (quoted-printable).
     '-Q' stands for 'binary zeroes', w/o trailing newline (quoted-printable).

   Notes:
   According to the XML specification, characters 0 through 31 and
   character 127 may not appare anywhere in adocument, not even
   as character references, such as &#31;. The only other possibility
   it to either remove them or turn them into something sensible.
   I took this second approach, whereby \001 (that is rc(1) list
   concatenation character) is replaced with a single blank. This is
   experimental, and may change in future versions of CSA.

   Beside escaping stuff, if this program is called by a CGI processor
   it also checks for a valid Content-Length value, and truncates
   extra input if any. This was introduced mainly to fix a wrong
   content-length value supplied by MSIE 5.0x.
 ****************************************************************************/

static char Psz_RCS[] = "$Date: 2004/08/08 21:15:38 $";

#include <stdio.h>
#include <stdlib.h>

static void usage() {
  fprintf(stderr,"usage: csa-escape [-p|-x|-i|-u|-a|-r|-s|-z|-Z]\n");
}

int main(int argc, char **argv) {

   int c, enc='p', nl=0, nbytes=0, clen=0;

   if ((argc > 1) &&
       (!strcmp(argv[1],"-h") || !strcmp(argv[1],"--help"))) {
      usage();
      exit(1);
   }

   if (argc > 1) {
      if (!strcmp(argv[1],"-p")) enc='p';
      else if (!strcmp(argv[1],"-x")) enc='x';
      else if (!strcmp(argv[1],"-i")) enc='i';
      else if (!strcmp(argv[1],"-u")) enc='u';
      else if (!strcmp(argv[1],"-r")) enc='r';
      else if (!strcmp(argv[1],"-s")) enc='s';
      else if (!strcmp(argv[1],"-a")) enc='a';
      else if (!strcmp(argv[1],"-z")) enc='z';
      else if (!strcmp(argv[1],"-Z")) enc='Z';
      else if (!strcmp(argv[1],"-q")) enc='q';
      else if (!strcmp(argv[1],"-Q")) enc='Q';
      else {
      	 usage();
      	 exit(1);
      }
   }

   /* Get HTTP content-length value if available */
   if (getenv("CONTENT_LENGTH") != NULL)
		clen = atoi(getenv("CONTENT_LENGTH"));

   if (enc == 'p') {

      /* PYX escaping: \ \t \n */

      putchar('-');

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

   	    if (nl) {
	        printf("\\n");
	        nl=0;
	    }

	    switch (c) {
	      case '\n':
	        nl=1;
	        break;
	      case '\t':
	        printf("\\t");
	        break;
	      case '\001':
	        putchar(' ');
	        break;
	      case '\\':
	        printf("\\\\");
	        break;
	      default:
   	        putchar(c);
	    }
      }

      putchar('\n');
   }
   else if (enc == 'x') {

      /* XML escaping: < " & 0x001 */

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

	    switch (c) {
	      case '<':
		printf("&lt;");
	        break;
	      case '"':
	        printf("&quot;");
	        break;
	      case '&':
		printf("&amp;");
	        break;
	      case '\001':
	        putchar(' ');
	        break;
	      default:
   	        putchar(c);
	    }
      }
   }
   else if (enc == 'i') {

      /* ISO escaping: & # < > | " ' ` \ \n \t \f SPACE */

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

	    switch (c) {
	      case '&':
		printf("&amp;");
	        break;
	      case '#':
		printf("&#35;");
	        break;
	      case '<':
		printf("&lt;");
	        break;
	      case '>':
		printf("&#62;");
	        break;
	      case '|':
		printf("&#124;");
	        break;
	      case '"':
	        printf("&#34;");
	        break;
	      case '\'':
		printf("&#39;");
	        break;
	      case '`':
		printf("&#96;");
	        break;
	      case '\n':
		printf("&#10;");
	        break;
	      case '\t':
		printf("&#9;");
	        break;
	      case '\f':
		printf("&#12;");
	        break;
	      case ' ':
		printf("&#32;");
	        break;
	      case '\\':
		printf("&#92;");
	        break;
	      default:
   	        putchar(c);
	    }
      }
   }
   else if (enc == 'u') {

      /* URI escaping:
	 % # & ? + SPACE = " \t \f ' \n \r $ ( ) [ ] ` < > \ / . |
	 ; ! * { } : ~ ,
      */

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

	    switch (c) {
	      case '%':
		printf("%%25");
	        break;
	      case '+':
	        printf("%%2B");
	        break;
	      case ' ':
		putchar('+');
	        break;
	      case '\t':
	        printf("%%09");
	        break;
	      case '\f':
	        printf("%%0C");
	        break;
	      case '\n':
	        printf("%%0A");
	        break;
	      case '\'':
	        printf("%%27");
	        break;
	      case '#':
	        printf("%%23");
	        break;
	      case '"':
	        printf("%%22");
	        break;
	      case '\r':
	        printf("%%0D");
	        break;
	      case '$':
	        printf("%%24");
	        break;
	      case '(':
	        printf("%%28");
	        break;
	      case ')':
	        printf("%%29");
	        break;
	      case '[':
	        printf("%%5B");
	        break;
	      case ']':
	        printf("%%5D");
	        break;
	      case '`':
	        printf("%%60");
	        break;
	      case '<':
	        printf("%%3C");
	        break;
	      case '>':
	        printf("%%3E");
	        break;
	      case '\\':
	        printf("%%5C");
	        break;
	      case '.':
	        printf("%%2E");
	        break;
	      case '|':
	        printf("%%7C");
	        break;
	      case ';':
	        printf("%%3B");
	        break;
	      case '!':
	        printf("%%21");
	        break;
	      case '*':
	        printf("%%2A");
	        break;
	      case '{':
	        printf("%%7B");
	        break;
	      case '}':
	        printf("%%7D");
	        break;
	      case ':':
	        printf("%%3A");
	        break;
	      case '~':
	        printf("%%7E");
	        break;
	      case ',':
	        printf("%%2C");
	        break;
	      /*
		A few characters need to be double-escaped, as they
		will be unescaped a first time by the CGI interface,
		and then a second time by the CSA library.
	      */
	      case '&':
	        printf("%%2526");
	        break;
	      case '?':
	        printf("%%253F");
	        break;
	      case '=':
	        printf("%%253D");
	        break;
	      case '/':
	        printf("%%252F");
	        break;
	      default:
   	        putchar(c);
	    }
      }
   }
   else if (enc == 'z' || enc == 'Z') {

      /* MAWK escaping: binary zeroes are turned into a '0' prefixed
	 with a backslash, and therefore the original backslashes
	 need to be scaped, too. Furthermore, unless the directive is 'Z',
	 an extra newline is added ad the end of the input stream, so
	 that a real ending newline, if any, will not be dropped by
	 urlencoded.awk. This problem cannot occur with multipart.awk,
	 due to the presence of the MIME part delimiter */

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

	    switch (c) {
	      case '\\':
		putchar('\\');
		putchar(c);
	        break;
	      case '\0':
		printf("\\%d",0);
	        break;
	      default:
   	        putchar(c);
	    }
      }
      if (enc == 'z') putchar('\n');
   }
   else if (enc == 'q' || enc == 'Q') {

      /* Same as z/Z, but encoding is done in quoted-printable */

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

	    switch (c) {
	      case '=':
		printf("=3D");
	        break;
	      case '\0':
		printf("=00");
	        break;
	      default:
   	        putchar(c);
	    }
      }
      if (enc == 'q') putchar('\n');
   }
   else if (enc == 'r') {

      /* rc(1) escaping: ' */

      putchar('\'');

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

   	    if (nl) {
	        printf("\\n");
	        nl=0;
	    }

	    switch (c) {
	      case '\n':
	        nl=1;
	        break;
	      case '\'':
		printf("''");
	        break;
	      default:
   	        putchar(c);
	    }
      }

      putchar('\'');

      if (nl) putchar('\n');
   }
   else if (enc == 's') {

      /* sh(1) escaping: \ " $ ` */

      putchar('"');

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

   	    if (nl) {
	        printf("\\n");
	        nl=0;
	    }

	    switch (c) {
	      case '\n':
	        nl=1;
	        break;
	      case '\\':
	        printf("\\\\");
	        break;
	      case '"':
	        printf("\\\"");
	        break;
	      case '$':
	        printf("\\$");
	        break;
	      case '`':
	        printf("\\`");
	        break;
	      default:
   	        putchar(c);
	    }
      }

      putchar('"');

      if (nl) putchar('\n');
   }
   else if (enc == 'a') {

      /* AWK escaping: \ " \n */

      putchar('"');

      while ((c = getchar()) != EOF) {

	    if (clen && ++nbytes > clen) exit(0);

   	    if (nl) {
	        printf("\\n");
	        nl=0;
	    }

	    switch (c) {
	      case '\n':
	        nl=1;
	        break;
	      case '\\':
	        printf("\\\\");
	        break;
	      case '"':
	        printf("\\\"");
	        break;
	      default:
   	        putchar(c);
	    }
      }

      putchar('"');

      if (nl) putchar('\n');
   }

   exit(0);
}

/* End of program */
